;Clock Timer Switch with 16 programmes and very flexible weekday assignments
;ATTiny24 @ 32768 Hz (ext. crystal)
;Fuses: low=0xC6, high=0xD6
;
;PORTS
;  PA0-3 = LCD Data 4-7 / Keys "Minute", "Hour", "Day", "Prog" (Min + Hr = DST)
;  PA4   = LCD RS / Key "Time"
;  PA5   = LCD E
;  PA6   = Key "On/Off"
;  PA7   = Key input (all keys)
;  PB2   = Relay Output (active high)

.equ	KEY_REPEAT_START = 40 ;delay until auto-repeat starts
.equ	KEY_REPEAT_INTERVAL = 15 ;delay between auto-repeats
	
;===============================================================================

.include "../include/tn24def.inc"

.def	sreg_backup = R0
.def	p_day = R1
.def	output = R2
.def	day_mask = R3

.def	temp = R16
.def	temp2 = R17
.def	count = R18
.def	second = R19
.def	minute = R20
.def	hour = R21
.def	day = R22
.def	keys = R23
.def	key_repeat = R24
.def	p_min = R25
.def	p_hour = R26
.def	p_num = R27
.def	sel_day = R28
.def	flags = R29
	.equ	fUpdateDisp = 0 ;update main display (time, day, output status)
	.equ	fProgModified = 1 ;selected programme modified
	.equ	fDST = 2 ;dayligt saving time
	.equ	fNoTimeSet = 3 ;time not set after power-on
;R31:R30 = Z pointer

.equ	LCD_PORT = PORTA
.equ	LCD_RS = 4
.equ	LCD_E = 5
.equ	KEY_PORT = PORTA

.equ	KEY_MIN = 0
.equ	KEY_HOUR = 1
.equ	KEY_DAY = 2
.equ	KEY_PROG = 3
.equ	KEY_TIME = 4
.equ	KEY_ONOFF = 6
.equ	KEY_PIN = PINA
.equ	KEY_RET = 7

.equ	OUT_PORT = PORTB
.equ	OUT_PIN = PINB
.equ	OUT_OUTPUT = 2

.equ	END = 0xFF ;string end marker

;===============================================================================

.cseg
.org 0x000
	rjmp	reset
.org OC0Aaddr
	rjmp	timer

;===============================================================================

timer:
	;timer 0 overflow interrupt: 1 Hz
	in	sreg_backup, SREG
	
	;no action if time is not set yet
	sbrc	flags, fNoTimeSet
	rjmp	timer_end
	
	;increment time
	inc	second
	cpi	second, 60
	brlo	timer_end
	clr	second
	ori	flags, 1<<fUpdateDisp
	inc	minute
	cpi	minute, 60
	brlo	timer_check
	clr	minute
	inc	hour
	cpi	hour, 24
	brlo	timer_check
	clr	hour
	inc	day
	cpi	day, 7
	brlo	timer_check
	clr	day
	
timer_check:
	;check programmes, switch output if match found
	push	temp
	push	temp2
	ldi	temp, 0x01 ;calculate weekday bitmask (Mo = 0x01, Su = 0x40)
	mov	day_mask, temp
	sbrc	day, 0
	lsl	day_mask
	sbrc	day, 1
	lsl	day_mask
	sbrc	day, 1
	lsl	day_mask
	sbrc	day, 2
	swap	day_mask
	ldi	temp, 0x04  ;set 'output' bit 2 to 1 for off, 0 for on
	sbis	OUT_PORT, OUT_OUTPUT
	ldi	temp, 0x00
	mov	output, temp
	ldi	temp2, 0
timer_check_loop:
	rcall	eeprom_read ;minute
	cp	temp, minute
	brne	timer_check_next
	rcall	eeprom_read ;hour
	cp	temp, hour
	brne	timer_check_next
	rcall	eeprom_read ;weekday
	and	temp, day_mask
	breq	timer_check_next
	mov	output, temp2 ;match found: move EEPROM address to 'output'
timer_check_next:
	andi	temp2, 0xFC ;calculate EEPROM address for next programme
	subi	temp2, -4
	cpi	temp2, 128 ;4 bytes * 16 programmes * 2 (on and off) = 128
	brlo	timer_check_loop
	sbrs	output, 2 ;Bit 2 of EEPROM address: 0 = on, 1 = off
	cbi	OUT_PORT, OUT_OUTPUT
	sbrc	output, 2
	sbi	OUT_PORT, OUT_OUTPUT
	pop	temp2
	pop	temp
timer_end:

	out	SREG, sreg_backup
	reti
	
;===============================================================================

reset:
	;set stack pointer
	ldi	temp, RAMEND
	out	SPL, temp
	
	;init ports
	ldi	temp, ~(1<<KEY_RET)
	out	DDRA, temp
	ldi	temp, 1<<KEY_RET
	out	PORTA, temp
	ldi	temp, 1<<OUT_OUTPUT
	out	DDRB, temp
	ldi	temp, 0x08 | 1<<OUT_OUTPUT ;Reset pullup
	out	PORTB, temp
	
	;disable timer 1, USI and ADC
	ldi	temp, 1<<PRTIM1 | PRUSI | PRADC
	out	PRR, temp
	
	;init timer 0
	ldi	temp, 1<<WGM01 ;CTC mode
	out	TCCR0A, temp
	ldi	temp, 0x04 ;Clk/256
	out	TCCR0B, temp
	ldi	temp, 127
	out	OCR0A, temp
	ldi	temp, 1<<OCIE0A
	out	TIMSK0, temp
	
	;init registers
	ldi	flags, 1<<fNoTimeSet
	clr	second
	clr	minute
	clr	hour
	clr	day
	clr	keys
	clr	key_repeat
	
	;init LCD
	rcall	lcd_init
	
	;display initial message
	ldi	temp, 0x40
	rcall	lcd_cmd
	rcall	lcd_str
		.db 0x08, 0x14, 0x14, 0x1C, 0x14, 0x14, 0x15, 0x00, END ;"A."
	ldi	temp, 0x80
	rcall	lcd_cmd
	rcall	lcd_str
		.db "Timer v1", END
	ldi	temp, 0xC0
	rcall	lcd_cmd
	rcall	lcd_str
		.db 0,"Rossius", END
	ldi	count, 8 ;delay 2 s
init_message_delay:
	ldi	temp, 250
	rcall	delay_ms
	dec	count
	brne	init_message_delay
	
	;LCD custom chars
	ldi	temp, 0x40
	rcall	lcd_cmd
	rcall	lcd_str
		.db 0x11, 0x1B, 0x15, 0x11, 0x11, 0x11, 0x00, 0x00 ;0: Monday
		.db 0x1F, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00 ;1: Tuesday
		.db 0x11, 0x11, 0x11, 0x15, 0x1B, 0x11, 0x00, 0x00 ;2: Wednesday
		.db 0x1F, 0x04, 0x04, 0x04, 0x04, 0x04, 0x00, 0x00 ;3: Thursday
		.db 0x1F, 0x10, 0x10, 0x1E, 0x10, 0x10, 0x00, 0x00 ;4: Friday
		.db 0x0E, 0x11, 0x0C, 0x06, 0x11, 0x0E, 0x00, 0x00 ;5: Saturday
		.db 0x0E, 0x11, 0x0C, 0x06, 0x11, 0x0E, 0x00, 0x00 ;6: Sunday
		.db 0x00, 0x00, 0x00, 0x0E, 0x11, 0x15, 0x0E, 0x0E ;7: lamp
		.db END
	
	;enable interrupts
	sei
	
;-------------------------------------------------------------------------------

main:
	;display main screen
	rcall	lcd_clear
	ldi	temp, 0x80
	rcall	lcd_cmd
	rcall	lcd_str
		.db 0, 1, 2, 3, 4, 5, 6, END ;days of the week
	ldi	temp, 0xC0
	rcall	lcd_cmd
	ldi	temp, 7 ;"lamp" symbol for on/off
	rcall	lcd_data
	ldi	temp, 0x0C ;no cursor
	rcall	lcd_cmd
	ori	flags, 1<<fUpdateDisp
	
loop:
	;update day, time and on/off display
	sbrc	flags, fUpdateDisp
	rcall	update_disp

	;on/off key: toggle output
	rcall	scan_keys
	sbrs	keys, KEY_ONOFF
	rjmp	key_onoff_end
	sbi	OUT_PIN, OUT_OUTPUT
	rcall	update_disp
key_onoff_loop:
	rcall	scan_keys
	sbrc	keys, KEY_ONOFF
	rjmp	key_onoff_loop
key_onoff_end:

	;prog key: enter programme menu
	rcall	scan_keys
	sbrc	keys, KEY_PROG
	rjmp	prog
	
	;set clock
	rcall	scan_keys
	sbrs	keys, KEY_TIME ;"Time" key must be pressed to set the clock
	rjmp	key_time_end
	andi	flags, ~(1<<fNoTimeSet)
	ori	flags, 1<<fUpdateDisp
	;Minutes
	sbrs	keys, KEY_MIN
	rjmp	key_time_min_end
key_time_min:
	clr	second
	inc	minute
	cpi	minute, 60
	brlo	PC+2
	clr	minute
	rcall	update_disp
	cpi	key_repeat, KEY_REPEAT_START
	brlo	PC+2
	subi	key_repeat, KEY_REPEAT_INTERVAL
key_time_min_loop:
	ldi	temp, 10
	rcall	delay_ms
	inc	key_repeat
	cpi	key_repeat, KEY_REPEAT_START
	breq	key_time_min
	rcall	scan_keys
	sbrs	keys, KEY_TIME
	rjmp	key_time_end
	sbrs	keys, KEY_MIN
	rjmp	key_time_end
	rjmp	key_time_min_loop
key_time_min_end:
	;Hours
	sbrs	keys, KEY_HOUR
	rjmp	key_time_hour_end
key_time_hour:
	inc	hour
	cpi	hour, 24
	brlo	PC+2
	clr	hour
	rcall	update_disp
	cpi	key_repeat, KEY_REPEAT_START
	brlo	PC+2
	subi	key_repeat, KEY_REPEAT_INTERVAL
key_time_hour_loop:
	ldi	temp, 10
	rcall	delay_ms
	inc	key_repeat
	cpi	key_repeat, KEY_REPEAT_START
	breq	key_time_hour
	rcall	scan_keys
	sbrs	keys, KEY_TIME
	rjmp	key_time_end
	sbrs	keys, KEY_HOUR
	rjmp	key_time_end
	rjmp	key_time_hour_loop
key_time_hour_end:
	;Days
	sbrs	keys, KEY_DAY
	rjmp	key_time_day_end
key_time_day:
	inc	day
	cpi	day, 7
	brlo	PC+2
	clr	day
	rcall	update_disp
	cpi	key_repeat, KEY_REPEAT_START
	brlo	PC+2
	subi	key_repeat, KEY_REPEAT_INTERVAL
key_time_day_loop:
	ldi	temp, 10
	rcall	delay_ms
	inc	key_repeat
	cpi	key_repeat, KEY_REPEAT_START
	breq	key_time_day
	rcall	scan_keys
	sbrs	keys, KEY_TIME
	rjmp	key_time_end
	sbrs	keys, KEY_DAY
	rjmp	key_time_end
	rjmp	key_time_day_loop
key_time_day_end:
key_time_end:
	clr	key_repeat
	
	;switch daylight saving time ("Min" + "Hour" keys both pressed)
	rcall	scan_keys
	sbrs	keys, KEY_MIN
	rjmp	key_dst_end
	sbrs	keys, KEY_HOUR
	rjmp	key_dst_end
	sbrc	flags, fDST
	rjmp	key_dst_disable
	;enable DST
	ori	flags, 1<<fDST
	inc	hour
	cpi	hour, 24
	brlo	key_dst_done
	clr	hour
	inc	day
	cpi	day, 7
	brlo	key_dst_done
	clr	day
	rjmp	key_dst_done
key_dst_disable:
	;disable DST
	andi	flags, ~(1<<fDST)
	dec	hour
	cpi	hour, 24
	brlo	key_dst_done
	ldi	hour, 23
	dec	day
	cpi	day, 7
	brlo	key_dst_done
	ldi	day, 6
key_dst_done:
	rcall	update_disp
key_dst_wait:
	rcall	scan_keys
	sbrs	keys, KEY_MIN
	rjmp	key_dst_end
	sbrc	keys, KEY_HOUR
	rjmp	key_dst_wait
key_dst_end:
	
	rjmp	loop

;===============================================================================

prog:
	;Programme Menu
	rcall	lcd_clear
	ldi	temp, 0x80
	rcall	lcd_cmd
	rcall	lcd_str
		.db 0, 1, 2, 3, 4, 5, 6, "a", END ;days
	ldi	temp, 0xC0
	rcall	lcd_cmd
	rcall	lcd_str
		.db "0", 7, " --:--", END
	ldi	temp, 0x0D ;enable blinking cursor
	rcall	lcd_cmd
	ldi	p_num, 0
	ldi	sel_day, 0
	andi	flags, ~(1<<fProgModified)
	
prog_next:
	;load programme from EEPROM
	rcall	prog_load
	;wait until prog key released
prog_next_wait:
	rcall	scan_keys
	sbrc	keys, KEY_PROG
	rjmp	prog_next_wait
	
prog_loop:
	;on/off key: toggle between "on" and "off" times for selected programme
	rcall	scan_keys
	sbrs	keys, KEY_ONOFF
	rjmp	prog_key_onoff_end
	sbrc	flags, fProgModified
	rcall	prog_save
	ldi	temp, 0x01
	eor	p_num, temp
	rcall	prog_load
prog_key_onoff_loop:
	rcall	scan_keys
	sbrc	keys, KEY_ONOFF
	rjmp	prog_key_onoff_loop
prog_key_onoff_end:

	;prog key: next programme
	rcall	scan_keys
	sbrs	keys, KEY_PROG
	rjmp	prog_key_prog_end
	sbrc	flags, fProgModified
	rcall	prog_save
	inc	p_num
	andi	p_num, 0x1F
	rjmp	prog_next
prog_key_prog_end:
	
	;time key: return to main display
	rcall	scan_keys
	sbrs	keys, KEY_TIME
	rjmp	prog_key_time_end
	sbrc	flags, fProgModified
	rcall	prog_save
	rjmp	main
prog_key_time_end:
	
	;Min/Hour keys: set time
	tst	p_day
	breq	prog_key_hour_end ;prevent setting time when "--:--" displayed
	rcall	scan_keys
	;Minutes
	sbrs	keys, KEY_MIN
	rjmp	prog_key_min_end
prog_key_min:
	ori	flags, 1<<fProgModified
	inc	p_min
	cpi	p_min, 60
	brlo	PC+2
	clr	p_min
	rcall	prog_update_disp
	cpi	key_repeat, KEY_REPEAT_START
	brlo	PC+2
	subi	key_repeat, KEY_REPEAT_INTERVAL
prog_key_min_loop:
	ldi	temp, 10
	rcall	delay_ms
	inc	key_repeat
	cpi	key_repeat, KEY_REPEAT_START
	breq	prog_key_min
	rcall	scan_keys
	sbrc	keys, KEY_MIN
	rjmp	prog_key_min_loop
prog_key_min_end:
	clr	key_repeat
	;Hours
	sbrs	keys, KEY_HOUR
	rjmp	prog_key_hour_end
prog_key_hour:
	ori	flags, 1<<fProgModified
	inc	p_hour
	cpi	p_hour, 24
	brlo	PC+2
	clr	p_hour
	rcall	prog_update_disp
	cpi	key_repeat, KEY_REPEAT_START
	brlo	PC+2
	subi	key_repeat, KEY_REPEAT_INTERVAL
prog_key_hour_loop:
	ldi	temp, 10
	rcall	delay_ms
	inc	key_repeat
	cpi	key_repeat, KEY_REPEAT_START
	breq	prog_key_hour
	rcall	scan_keys
	sbrc	keys, KEY_HOUR
	rjmp	prog_key_hour_loop
prog_key_hour_end:
	clr	key_repeat
	
	;Day
	rcall	scan_keys
	sbrs	keys, KEY_DAY
	rjmp	prog_key_day_end
prog_key_day:
	ldi	temp, 10
	rcall	delay_ms
	inc	key_repeat
	cpi	key_repeat, KEY_REPEAT_START
	brsh	prog_key_day_long
	rcall	scan_keys
	sbrc	keys, KEY_DAY
	rjmp	prog_key_day
	
	;short keypress: select day
	inc	sel_day
	cpi	sel_day, 8
	brlo	PC+2
	ldi	sel_day, 0
	rcall	prog_update_disp
	rjmp	prog_key_day_end

prog_key_day_long:
	;long keypress: toggle active/not active for selected day
	ori	flags, 1<<fProgModified
	ldi	temp, 0x01
	sbrc	sel_day, 0
	lsl	temp
	sbrc	sel_day, 1
	lsl	temp
	sbrc	sel_day, 1
	lsl	temp
	sbrc	sel_day, 2
	swap	temp
	cpi	sel_day, 7 ;selected "a" (all enabled/disabled)
	brne	prog_key_day_all_end
	ldi	temp, 0x00 ;for 'eor' => no change
	ldi	temp2, 0x7F
	cp	p_day, temp2
	breq	prog_key_day_all_off
	mov	p_day, temp2
	rjmp	prog_key_day_all_end
prog_key_day_all_off:
	clr	p_day
prog_key_day_all_end:
	eor	p_day, temp
	rcall	prog_update_disp
prog_key_day_loop:
	rcall	scan_keys
	sbrc	keys, KEY_DAY
	rjmp	prog_key_day_loop
	
prog_key_day_end:
	clr	key_repeat
	
	
	rjmp	prog_loop




;===============================================================================

scan_keys:
	;scan keys, debounce if changed, return which keys are pressed down
	mov	temp2, keys
	clr	keys
	ldi	temp, ~(1<<KEY_MIN | 1<<LCD_E) ;"Minute"
	out	KEY_PORT, temp
	rjmp	PC+1
	sbis	KEY_PIN, KEY_RET
	ori	keys, 1<<KEY_MIN
	ldi	temp, ~(1<<KEY_HOUR | 1<<LCD_E) ;"Hour"
	out	KEY_PORT, temp
	rjmp	PC+1
	sbis	KEY_PIN, KEY_RET
	ori	keys, 1<<KEY_HOUR
	ldi	temp, ~(1<<KEY_DAY | 1<<LCD_E) ;"Day"
	out	KEY_PORT, temp
	rjmp	PC+1
	sbis	KEY_PIN, KEY_RET
	ori	keys, 1<<KEY_DAY
	ldi	temp, ~(1<<KEY_PROG | 1<<LCD_E) ;"Prog"
	out	KEY_PORT, temp
	rjmp	PC+1
	sbis	KEY_PIN, KEY_RET
	ori	keys, 1<<KEY_PROG
	ldi	temp, ~(1<<KEY_TIME | 1<<LCD_E) ;"Time"
	out	KEY_PORT, temp
	rjmp	PC+1
	sbis	KEY_PIN, KEY_RET
	ori	keys, 1<<KEY_TIME
	ldi	temp, ~(1<<KEY_ONOFF | 1<<LCD_E) ;"On/Off"
	out	KEY_PORT, temp
	rjmp	PC+1
	sbis	KEY_PIN, KEY_RET
	ori	keys, 1<<KEY_ONOFF
	eor	temp2, keys
	breq	scan_keys_end
	ldi	temp, 20 ;debounce
	rcall	delay_ms
scan_keys_end:
	ret
	
;--------------------

update_disp:
	;update main screen display (day, time and on/off status)
	
	;disable interrupts and copy day/hour/minute
	;to prevent race condition, e.g.:
	;1) display day
	;2) display hour
	;<interrupt: Mo 23:59 => Tu 00:00>
	;3) display minute
	;  ==> resulting display: "Mo 23:00"
	cli
	mov	p_min, minute
	mov	p_hour, hour
	mov	p_day, day
	andi	flags, ~(1<<fUpdateDisp)
	sei
	
	ldi	count, 0
update_disp_day:
	;day of the week
	ldi	temp, 0x47
	mov	temp2, count
	swap	temp2
	lsr	temp2
	add	temp, temp2
	rcall	lcd_cmd
	ldi	temp, 0x00
	cp	count, p_day
	brne	PC+2
	ldi	temp, 0x1F
	rcall	lcd_data
	inc	count
	cpi	count, 7
	brlo	update_disp_day
	;time
	ldi	temp, 0xC2
	rcall	lcd_cmd
	sbrs	flags, fNoTimeSet
	rjmp	update_disp_time
	rcall	lcd_str
		.db "TIME?", END
	rjmp	update_disp_time_end
update_disp_time:
	ldi	temp, ' ' ;DST disabled
	sbrc	flags, fDST
	ldi	temp, '+' ;DST enabled
	rcall	lcd_data
	mov	temp2, p_hour
	rcall	bin2lcd
	ldi	temp, ':'
	rcall	lcd_data
	mov	temp2, p_min
	rcall	bin2lcd
update_disp_time_end:
	;output on/off
	ldi	temp, 0x78
	rcall	lcd_cmd
	ldi	temp, 0x00
	sbis	OUT_PORT, OUT_OUTPUT
	ldi	temp, 0x15
	push	temp
	rcall	lcd_data
	pop	temp
	rcall	lcd_data
	ret
	
;--------------------
	
prog_save:
	;save programme to EEPROM
	mov	temp2, p_num
	lsl	temp2
	lsl	temp2
	mov	temp, p_min
	rcall	eeprom_write
	mov	temp, p_hour
	rcall	eeprom_write
	mov	temp, p_day
	rcall	eeprom_write
	andi	flags, ~(1<<fProgModified)
	ret

;--------------------

prog_load:
	;load programme from EEPROM
	mov	temp2, p_num
	lsl	temp2
	lsl	temp2
	rcall	eeprom_read
	cpi	temp, 60
	brlo	PC+3
	ldi	temp, 0
	ori	flags, 1<<fProgModified
	mov	p_min, temp
	rcall	eeprom_read
	cpi	temp, 24
	brlo	PC+3
	ldi	temp, 0
	ori	flags, 1<<fProgModified
	mov	p_hour, temp
	rcall	eeprom_read
	cpi	temp, 0x80
	brlo	PC+3
	ldi	temp, 0x00 ;bitmask for days => 0x00 = programme disabled
	ori	flags, 1<<fProgModified
	mov	p_day, temp
	;(fall through to prog_update_disp)

;--------------------

prog_update_disp:
	;update prog screen display (day, time, on/off and programme number)
	ldi	count, 0
prog_update_disp_day:
	;day(s) of the week
	ldi	temp, 0x47
	mov	temp2, count
	lsl	temp2
	lsl	temp2
	lsl	temp2
	add	temp, temp2
	rcall	lcd_cmd
	ldi	temp2, 0x01
	sbrc	count, 0
	lsl	temp2
	sbrc	count, 1
	lsl	temp2
	sbrc	count, 1
	lsl	temp2
	sbrc	count, 2
	swap	temp2
	ldi	temp, 0x00
	and	temp2, p_day
	breq	PC+2
	ldi	temp, 0x1F
	rcall	lcd_data
	inc	count
	cpi	count, 7
	brlo	prog_update_disp_day
	;programme no.
	ldi	temp, 0xC0
	rcall	lcd_cmd
	mov	temp, p_num
	lsr	temp
	subi	temp, -'0'
	cpi	temp, '9'+1
	brlo	PC+2
	subi	temp, -7
	rcall	lcd_data
	;time
	ldi	temp, 0xC3
	rcall	lcd_cmd
	tst	p_day
	breq	prog_update_disp_disabled ;programme disabled: display "--:--"
	mov	temp2, p_hour
	rcall	bin2lcd
	ldi	temp, ':'
	rcall	lcd_data
	mov	temp2, p_min
	rcall	bin2lcd
prog_update_disp_time_end:
	;output on/off
	ldi	temp, 0x78
	rcall	lcd_cmd
	ldi	temp, 0x00
	sbrs	p_num, 0 ;even numbers (0, 2, ...) = on, odd (1, 3, ...) = off
	ldi	temp, 0x15
	push	temp
	rcall	lcd_data
	pop	temp
	rcall	lcd_data
	;set cursor to selected day
	ldi	temp, 0x80
	add	temp, sel_day
	rcall	lcd_cmd
	ret
	
prog_update_disp_disabled:
	rcall	lcd_str
		.db "--:--", END
	rjmp	prog_update_disp_time_end
;===============================================================================

lcd_init:
	ldi	temp, 50 ;delay 50 ms
	rcall	delay_ms
	ldi	temp, 0x03 | 1<<KEY_RET
	rcall	lcd_write_8
	ldi	temp, 1 ;delay 1 ms
	rcall	delay_ms
	ldi	temp, 0x03 | 1<<KEY_RET
	rcall	lcd_write_8
	ldi	temp, 0x03 | 1<<KEY_RET
	rcall	lcd_write_8
	ldi	temp, 0x02 | 1<<KEY_RET ;Function Set: 4 bit
	rcall	lcd_write_8
	ldi	temp, 0x2A ;Function Set: 4 bit, 2 lines, 5x7 font
	rcall	lcd_cmd
	ldi	temp, 0x0C ;Display on/off: on, no cursor
	rcall	lcd_cmd
	ldi	temp, 0x06 ;Entry Mode Set: cursor increment, no shift
	rcall	lcd_cmd
	ldi	temp, 0x02 ;Return Home
	rcall	lcd_cmd
	ldi	temp, 2 ;delay 2 ms
	rcall	delay_ms
lcd_clear:
	ldi	temp, 0x01 ;Clear Display
	rcall	lcd_cmd
	ldi	temp, 2 ;delay 2 ms
	rcall	delay_ms
	ret
	
;--------------------

bin2lcd:
	;display binary number 'temp2' on the LCD (2 digits, 00-99)
	ldi	temp, '0'
bin2lcd_conv:
	cpi	temp2, 10
	brlo	bin2lcd_conv_end
	subi	temp2, 10
	inc	temp
	rjmp	bin2lcd_conv
bin2lcd_conv_end:
	ori	temp2, '0'
	push	temp2
	rcall	lcd_data
	pop	temp
	rcall	lcd_data
	ret

;--------------------

lcd_str:
	pop	ZH
	pop	ZL
	lsl	ZL
	rol	ZH
lcd_str_loop:
	lpm	temp, Z+
	cpi	temp, END
	breq	lcd_str_end
	rcall	lcd_data
	rjmp	lcd_str_loop
lcd_str_end:
	adiw	ZH:ZL, 1
	lsr	ZH
	ror	ZL
	ijmp ;return

;--------------------

lcd_cmd:
	mov	temp2, temp
	swap	temp2
	andi	temp2, 0x0F
	ori	temp2, 1<<KEY_RET
	andi	temp, 0x0F
	ori	temp, 1<<KEY_RET
	rjmp	lcd_write
	
;--------------------

lcd_data:
	mov	temp2, temp
	swap	temp2
	andi	temp2, 0x0F
	ori	temp2, 1<<LCD_RS | 1<<KEY_RET
	andi	temp, 0x0F
	ori	temp, 1<<LCD_RS | 1<<KEY_RET
lcd_write:
	out	LCD_PORT, temp2
	sbi	LCD_PORT, LCD_E
	cbi	LCD_PORT, LCD_E
lcd_write_8:
	out	LCD_PORT, temp
	sbi	LCD_PORT, LCD_E
	cbi	LCD_PORT, LCD_E
	ret

;===============================================================================

eeprom_write:
	;write one byte to EEPROM, increment address
	cli
	sbic	EECR, EEPE
	rjmp	eeprom_write
	out	EEARL, temp2
	out	EEDR, temp
	sbi	EECR, EEMPE
	sbi	EECR, EEPE
	inc	temp2
	sei
	ret
	
;--------------------

eeprom_read:
	;read one byte from EEPROM, increment address
	cli
	sbic	EECR, EEPE
	rjmp	eeprom_read
	out	EEARL, temp2
	sbi	EECR, EERE
	in	temp, EEDR
	inc	temp2
	sei
	ret

;===============================================================================

delay_ms:
	;temp * ~1 ms delay
	ldi	temp2, 10
delay_ms_loop:
	dec	temp2
	brne	delay_ms_loop
	dec	temp
	brne	delay_ms
	ret
	